export class MidiPlayer {
    constructor({ encodeMidiMessage, filterMidiMessage, json, midiFileSlicer, midiOutput, scheduler }) {
        this._encodeMidiMessage = encodeMidiMessage;
        this._filterMidiMessage = filterMidiMessage;
        this._json = json;
        this._midiFileSlicer = midiFileSlicer;
        this._midiOutput = midiOutput;
        this._scheduler = scheduler;
        this._state = null;
    }
    play() {
        if (this._state !== null) {
            throw new Error('The player is currently playing.');
        }
        return new Promise((resolve, reject) => {
            const schedulerSubscription = this._scheduler.subscribe({
                error: (err) => reject(err),
                next: ({ end, start }) => {
                    if (this._state === null) {
                        this._state = { endedTracks: 0, offset: start, resolve, schedulerSubscription: null };
                    }
                    this._schedule(start, end, this._state);
                }
            });
            if (this._state === null) {
                schedulerSubscription.unsubscribe();
            }
            else {
                this._state.schedulerSubscription = schedulerSubscription;
            }
        });
    }
    stop() {
        if (this._state === null) {
            throw new Error('The player is already stopped.');
        }
        this._stop(this._state);
    }
    _schedule(start, end, state) {
        const events = this._midiFileSlicer.slice(start - state.offset, end - state.offset);
        events
            .filter(({ event }) => this._filterMidiMessage(event))
            .forEach(({ event, time }) => this._midiOutput.send(this._encodeMidiMessage(event), start + time));
        const endedTracks = events.filter(({ event }) => MidiPlayer._isEndOfTrack(event)).length;
        state.endedTracks += endedTracks;
        if (state.endedTracks === this._json.tracks.length) {
            this._stop(state);
        }
    }
    _stop(state) {
        const { resolve, schedulerSubscription } = state;
        schedulerSubscription === null || schedulerSubscription === void 0 ? void 0 : schedulerSubscription.unsubscribe();
        this._state = null;
        resolve();
    }
    static _isEndOfTrack(event) {
        return 'endOfTrack' in event;
    }
}
//# sourceMappingURL=midi-player.js.map